#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "structures.h"
#include "maillage_local.h"
#include "etend_maillage.h"

/* Pour eviter le exit(-1); */
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include "xdebug.h"

extern int compteur_arret;
extern maillage *maillage_demo;

pt3d *pt_front ( pt3d *begin, maillage *maillage)
{
    unsigned int i;
    for (i=begin-maillage->first;i<maillage->nb;i++)
	if (maillage->maille[i].voisin != NULL)
	    return (maillage->first + i);
    return NULL;
}


unsigned int etend_maillage (	pt3d *point,
				vect *normale,
				maillage *maillage, 
				str_mem *tab)
{
    unsigned int i, j, k, k2, flag, t;
    unsigned int id_voisin, id_prec, id_foll, front, swap, join;
    unsigned short flag_front = 0;
    unsigned short flag_swap = 0;
    unsigned short flag_join = 0;
    unsigned int total = 0;
    pt2d *project = projete( normale, tab);
    pt3d *amax = NULL;
    pt3d *bmax = NULL;
    pt3d *pmax = NULL;
    pt3d *qmax = NULL;
    pt3d *rmax = NULL;
    pt3d *voisin, *foll, *prec, *pti, *ptj, *ptt;
    double temp0, temp1, temp2, temp3, medmax, taille;
    liste **liste_elt, *liste_ptr;

    /*
     * Trouver un point non maille, ou sur la frontiere, 
     * et un segment de la frontiere, qui peuvent etre
     * mailles au mieux.
     *
     * Parcours bourrin de toutes les possibilites:
     * pour chaque segment de la frontiere, je cherche le point
     * le mieux adapte. Ensuite, je prends le meilleur triplet, 
     * J'update le maillage, et je recommence...
     */

    while (compteur_arret > 0)
/* Ca c'est pour le debug => en realite: while (TRUE). */
      {
	medmax = DBL_MAX;
/* La limite 0.9, c'est pour se fixer une limite inferieure
 * de triangle acceptable.
 */
    for (i=0; (i<tab->current)/* && (medmax > 0.9) */; i++)
    {
	pti = tab->debut[i];
	if ( ((voisin = maillage->maille[pti-maillage->first].voisin) != NULL) &&
	     (maillage->maille[pti-maillage->first].voisins != NULL) )
	{
	    /*
	     * i est sur la frontiere.
	     * On a le segment par celui qui
	     * est pointe, si ca ne sort pas...
	     * Il faut aussi que les points consecutifs au segment,
	     * sur la frontiere, soient dans la zone locale,
	     * pour verifier qu'il n'y a pas de croisement.
	     */

	    id_voisin = 0;
	    while ((id_voisin < tab->current) && (voisin != tab->debut[id_voisin]))
		id_voisin++;

	    id_prec = 0;
	    while ( (id_prec < tab->current) &&
		    (maillage->maille[tab->debut[id_prec]-maillage->first].voisin != pti))
		id_prec++;
	    prec = tab->debut[id_prec];

	    foll = maillage->maille[voisin-maillage->first].voisin;
	    id_foll = 0;
	    while ( (id_foll < tab->current) &&
		    (foll != tab->debut[id_foll]))
		id_foll++;
		
	    if ((id_voisin < tab->current) &&
		(id_prec   < tab->current) &&
		(id_foll   < tab->current))
	    {
		/*
		 * On a un segment, reste plus qu'a trouver le meilleur point.
		 * Ensuite, si cette paire est meilleure, on update (*max).
		 */

		for (j=0; j<tab->current; j++)
		{
		    /*
		     * Faut deja qu'il soit different des 2 extremites...
		     * Et pas du tout maille, ou bien sur la frontiere,
		     * consecutif a une des 2 extremites (on se limite a
		     * un seul cas, c'est symetrique).
		     * En consequence, la frontiere reste toujours connexe.
		     * C'est plus simple a gerer.
		     */
		    
		    ptj = tab->debut[j];

		    if (maillage->maille[ptj - maillage->first].voisin != NULL) {
			front = (j == id_foll);
			join  = (!front ) && (j != id_voisin) && (j != i) && (j != id_prec);
		    }
		    else {
			front = join = FALSE;
		    };

		    if (join) {
			ptt = maillage->maille[ptj-maillage->first].voisin;
			t = 0;
			while ( (t < tab->current) && (tab->debut[t] != ptt) ) t++;
			if (t < tab->current)
			  {
			    /* Test pour verifier que "a" et "p" ne sont pas relies => ca pourrait defaire !!! */
			    liste_ptr = maillage->maille[ptj-maillage->first].voisins;
			    while ((liste_ptr != NULL) && (liste_ptr->point != pti))
			      liste_ptr = liste_ptr->next;
			    join = (liste_ptr == NULL);
			  }
			else join = FALSE;
		    };
		    
		  if ( (maillage->maille[ptj - maillage->first].voisins == NULL) ||
		       (front) || (join) )
		    {
			flag = inter_check ( maillage, tab, project,
					     pti, ptj, voisin,
					     i, j, id_voisin);

			if (join)
			    flag =  flag &&
				    inter_check (   maillage, tab, project,
						    pti, ptj, ptt,
						    i, j, t) &&
				    OPPSIDE(project[i], project[j], project[t], project[id_voisin]) &&
				    ( (maillage->maille[foll - maillage->first].voisin != ptj) ||
				      (inter_check (	maillage, tab, project,
							voisin, ptj, foll,
							id_voisin, j, id_foll))) &&
				    ( (maillage->maille[ptt - maillage->first].voisin != prec) ||
				      (inter_check (	maillage, tab, project,
							ptt, pti, prec,
							t, i, id_prec)));

			/*
			 * On ne garde que celui qui est "optimal".
			 * Ca donne un triangle potable.
			 * Il faudrait eventuellement tenir compte du futur swap...
			 */

			taille = MED_ABS( project[i], project[id_voisin], project[j]);

			if ( (flag) && (taille < medmax) )
			{
			    /*
			     * Il faut qu'il soit du bon cote...
			     * => determiner le BON (!) voisin commun a amax et bmax.
			     * (de toute facon, on en a besoin pour l'optimisation)
			     * 
			     * Rq: dans le cas d'un point sur la frontiere,
			     *     ca peut-etre le meme... => meme cote =>
			     *     donc ca le vire.
			     */

			    k = bon_voisin (maillage, tab, pti, voisin);

			    /* Si p est le suivant de b sur la frontiere, on cherche
			     * leur bon voisin commun => rmax, pour un swap eventuel. */

			    if (front)
				k2 = bon_voisin (maillage, tab, ptj, voisin);

				/* On verifie que p et q sont sur les cotes opposes. */

			    if ( (k<tab->current) &&
				 OPPSIDE(project[i], project[id_voisin], project[j], project[k]) )
			    {
				/*
				 * Le point est retenu.
				 *
				 * swap = 0 pour pas de swap,
				 *        1 pour swap de base,
				 *     et 2 pour swap avec rmax.
				 */
				 
			        swap = 0;
			        amax = pti;
			        bmax = voisin;
			        pmax = ptj;
				qmax = tab->debut[k]; 
			        flag_front = front;
				medmax = taille;
				temp0 = MAX(medmax, MED_ABS(project[i], project[k], project[id_voisin]));
				temp1 = MAX(MED_ABS(project[j], project[i], project[k]),
					    MED_ABS(project[j], project[id_voisin], project[k]));
				if ( (temp0 > temp1) &&
				     OPPSIDE(project[j], project[k], project[i], project[id_voisin]) )
				    swap = 1;

				if ( (front) && (k2<tab->current) &&
				     OPPSIDE(project[i], project[k2], project[j], project[id_voisin]) )
				{
				    temp2 = MAX( medmax,
						 MED_ABS(project[j], project[k2], project[id_voisin]));
				    temp3 = MAX( MED_ABS(project[j], project[i], project[k2]),
				    		 MED_ABS(project[i], project[id_voisin], project[k2]));
				    if ((temp2 > temp3) && (MIN(temp0,temp1) > temp2))
				    {
				        swap = 2;
				        rmax = tab->debut[k2];
				    };
				}
				else if (join)
				{
/* A noter : il faudrait faire un check symetrique de "p et q" sont du cote oppose. */
				    swap = 0;
				    rmax = ptt;
				};
				flag_swap = swap;				    
				flag_join = join;
			    };
			};
		    };
		};
	    };
	};
    };
    
/* On a donc a mailler amax, bmax, pmax, [qmax, rmax]. */
/* => test pour verifier qu'on a trouve quelque chose. */

    if (medmax == DBL_MAX)
    {
	free (project);
	return total;
    }
    else
    {
	++total;
	
      /* On update la frontiere. */

      if (flag_front) {
	/* Cas ou a->b->p devient a->p. (l'autre cas a ete elimine) */
	maillage->maille[amax - maillage->first].voisin = pmax;
	maillage->maille[bmax - maillage->first].voisin = NULL;
      } else if (flag_join) {
	maillage->maille[pmax - maillage->first].voisin = bmax;
	if (maillage->maille[rmax - maillage->first].voisin == amax)
	{
	  maillage->maille[amax - maillage->first].voisin = NULL;
	  maillage->maille[rmax - maillage->first].voisin = NULL;
	} else {
	  maillage->maille[amax - maillage->first].voisin = rmax;
	};
      } else {
	maillage->maille[amax - maillage->first].voisin = pmax;
	maillage->maille[pmax - maillage->first].voisin = bmax;
      };
      
      if (flag_join)
      {
	/* Il faut rajouter 3 liaisons: (p<->b),  (p<->a) et (a<->r). */
/* b */
	insert_liste(	&maillage->maille[bmax-maillage->first].voisins,
			amax, pmax, maillage->maille[bmax-maillage->first].voisin);
/* a */
	if (maillage->maille[amax - maillage->first].voisin != NULL)
	{
	    liste_elt = &(maillage->maille[amax - maillage->first].voisins);
	    while (maillage->maille[(*liste_elt)->point - maillage->first].voisin != amax)
		liste_elt = &((*liste_elt)->next);
	    insert_liste(   &maillage->maille[amax-maillage->first].voisins,
			    bmax, rmax, (*liste_elt)->point);
	};
	insert_liste(	&maillage->maille[amax-maillage->first].voisins,
			bmax, pmax, rmax);
/* r */
	if (maillage->maille[amax - maillage->first].voisin != NULL)
	{
	    insert_liste(   &maillage->maille[rmax-maillage->first].voisins,
			    pmax, amax, maillage->maille[rmax-maillage->first].voisin);
	};
/* p */
        liste_elt = &(maillage->maille[pmax - maillage->first].voisins);
	while (maillage->maille[(*liste_elt)->point - maillage->first].voisin != pmax)
	    liste_elt = &((*liste_elt)->next);
	insert_liste(	&maillage->maille[pmax-maillage->first].voisins,
			rmax, bmax, (*liste_elt)->point);
	insert_liste(	&maillage->maille[pmax-maillage->first].voisins,
			bmax, amax, rmax);

/* On elimine les boucles de 3. */
	if (maillage->maille[maillage->maille[bmax - maillage->first].voisin - maillage->first].voisin == pmax)
	{
	    maillage->maille[maillage->maille[bmax - maillage->first].voisin - maillage->first].voisin = NULL;
	    maillage->maille[bmax - maillage->first].voisin = NULL;
	    maillage->maille[pmax - maillage->first].voisin = NULL;
	};
	if ( (maillage->maille[amax - maillage->first].voisin != NULL) &&
	     (maillage->maille[maillage->maille[rmax - maillage->first].voisin - maillage->first].voisin == amax))
	{
	    maillage->maille[maillage->maille[rmax - maillage->first].voisin - maillage->first].voisin = NULL;
	    maillage->maille[rmax - maillage->first].voisin = NULL;
	    maillage->maille[amax - maillage->first].voisin = NULL;
	};
      }
      else if (flag_swap == 0)
	{   /* Pas de swap. */

	    /* On insere pmax dans la liste de amax,
	     * du cote oppose a qmax, par rapport a bmax.
	     */

    	    croise_liste ( &(maillage->maille[amax - maillage->first].voisins),
			   pmax, bmax, qmax);
	    
	    if (!flag_front)
	    {
		/* On insere pmax dans la liste de bmax.
		 * Du cote oppose a qmax par rapport a amax.
		 */

	    	croise_liste (	&(maillage->maille[bmax - maillage->first].voisins),
				pmax, amax, qmax);

		/* On insere amax et bmax dans la liste de pmax. */
		NEW_LISTE(liste_ptr);
		maillage->maille[pmax - maillage->first].voisins = liste_ptr;
		liste_ptr->point = bmax;
		NEW_LISTE(liste_ptr->next);
		liste_ptr->next->point = amax;
		liste_ptr->next->next  = NULL;
	    }
	    else
	    {
		/* Cas ou b est deja dans la liste, on insere a entre b,
		 * et un voisin de p, son suivant sur la frontiere.
		 */

		insert_liste( &(maillage->maille[pmax - maillage->first].voisins), bmax,
			      amax, maillage->maille[pmax-maillage->first].voisin);
	    };
    	}
	else if (flag_swap == 1)
	{
	    /* On swap: (ab) => (pq). */

	     /*
	      * En detail, il faut remplacer a<->b par a<->p et p<->b.
	      * En inserant dans la liste de p (q<->p), et rajouter p
	      * dans la liste de q, entre a et b.
	      */

	    /* Je fais/modifie la liste de p. */

            if (!flag_front)
	      {
		NEW_LISTE(liste_ptr);
		maillage->maille[pmax - maillage->first].voisins = liste_ptr;
		liste_ptr->point = bmax;
		NEW_LISTE(liste_ptr->next);
		liste_ptr->next->point = qmax;
		NEW_LISTE(liste_ptr->next->next);
		liste_ptr->next->next->point = amax;
		liste_ptr->next->next->next  = NULL;

		/* Je modifie la liste de a (remplacer b par p). */
		replace_liste( &maillage->maille[amax - maillage->first].voisins, bmax, pmax);

		/* Je modifie la liste de b (remplacer a par p). */
		replace_liste( &maillage->maille[bmax - maillage->first].voisins, amax, pmax);
	      }
	    else
	      {
		/* Rajouter a puis q entre b et le suivant de p sur la frontiere. */
		insert_liste (  &(maillage->maille[pmax - maillage->first].voisins),
				bmax, amax, maillage->maille[pmax - maillage->first].voisin);
		insert_liste (	&(maillage->maille[pmax - maillage->first].voisins),
				amax, qmax, bmax);

		/* Je modifie la liste de b (remplacer a par p). */
		delete_liste(  &maillage->maille[bmax - maillage->first].voisins, amax);

		/* Je modifie la liste de a (remplacer b par p). */
		replace_liste( &maillage->maille[amax - maillage->first].voisins, bmax, pmax);
	      };


	    /* Je modifie la liste de q (inserer p entre a et b). */
	    insert_liste (&(maillage->maille[qmax - maillage->first].voisins),
			  amax, pmax, bmax);
	}
	else if (flag_swap == 2)
	{
	    /* On swap: (pb) => (ar). */

	     /*
	      * En detail, il faut remplacer p<->b par a<->p et a<->r.
	      * En inserant dans la liste de a (a<->r) et (a<->p),
	      * et rajouter a dans la liste de r, entre p et b.
	      */

	    /* Je modifie la liste de a: */
	    /* Rajouter p puis r entre b et le precedent de a sur la frontiere. */

	    liste_elt = &(maillage->maille[amax - maillage->first].voisins);
	    while (maillage->maille[(*liste_elt)->point - maillage->first].voisin != amax)
		liste_elt = &((*liste_elt)->next);

	    insert_liste (  &(maillage->maille[amax - maillage->first].voisins),
	    		    (*liste_elt)->point, pmax, bmax);
	    insert_liste (  &(maillage->maille[amax - maillage->first].voisins),
	    		    pmax, rmax, bmax);

	    /* Je modifie la liste de p. (remplacer b par a) */
	    replace_liste( &(maillage->maille[pmax - maillage->first].voisins), bmax, amax);

	    /* Je modifie la liste de b (eliminer p). */
	    delete_liste( &(maillage->maille[bmax - maillage->first].voisins), pmax);

	    /* Je modifie la liste de r (inserer a entre p et b). */
	    insert_liste( &(maillage->maille[rmax - maillage->first].voisins), pmax, amax, bmax);
	};

    if ( (flag_front) &&
	 (maillage->maille[maillage->maille[maillage->maille[pmax-maillage->first].voisin-maillage->first].voisin-maillage->first].voisin == pmax) )
    {
	maillage->maille[maillage->maille[maillage->maille[pmax-maillage->first].voisin-maillage->first].voisin-maillage->first].voisin = NULL;
	maillage->maille[maillage->maille[pmax-maillage->first].voisin-maillage->first].voisin = NULL;
	maillage->maille[pmax-maillage->first].voisin = NULL;
    };

/* Ne sert que pour le debug: */
if (compteur_arret > 0) --compteur_arret;
    };
      };
printf("Ne doit s'afficher qu'en mode debug.\n");
    return total;
}


/*
 * Routine d'insertion de p dans la liste pointee
 * par liste_p, entre a et b.
 */

void insert_liste (liste **liste_p, pt3d *a, pt3d *p, pt3d *b)
{
  liste **liste_elt = liste_p;
  liste *liste_ptr;
  NEW_LISTE(liste_ptr);
  liste_ptr->point = p;
  
  while ( ((*liste_elt)->point != a) && ((*liste_elt)->point != b))
    liste_elt = &((*liste_elt)->next);

if ((*liste_elt)->next == NULL)
{
    /* Juste pour le debug. */
    
/* Cette erreur est plus subtile qu'on pourrait croire.
 * Elle peut arriver si le maillage est inconsistant trivialement,
 * ou bien si l'inconsistance est due a une insertion dont un des
 * 2 voisins n'est pas dans la zone locale. Enfin je crois.
 */
printf("Insertion de p entre a et b => y'a qu'un seul des elements.\n");
x_dessine(maillage_demo); /* exit(-1); */
};

  if ( ((*liste_elt)->next->point == a) ||
       ((*liste_elt)->next->point == b) )
    {
      /* a et b sont consecutifs: insertion bete. */
      liste_ptr->next = (*liste_elt)->next;
      (*liste_elt)->next = liste_ptr;
    }
  else
    {
      /*
       * Cas ou l'un est en derniere position, l'autre en premiere.
       * On integre p en premiere.
       */
if (liste_p != liste_elt) 
{
printf("Erreur d'insertion.\n");
};
     liste_ptr->next    = (*liste_p);
      (*liste_p) = liste_ptr;
    };
}


/*
 * Routine de remplacement de a par p, dans la liste de b.
 */

void replace_liste (liste **liste_elt, pt3d *a, pt3d *p)
{
  while ((*liste_elt)->point != a)
    liste_elt = &((*liste_elt)->next);
  (*liste_elt)->point = p;
}


/*
 * Routine de deletion de a dans la liste de b.
 */

void delete_liste (liste **liste_elt, pt3d *a)
{
  liste *interim;
  while ((*liste_elt)->point != a)
    liste_elt = &((*liste_elt)->next);

  /* On perd un peu de temps ici a faire ce check => mettre en debug */

if ((*liste_elt) == NULL)
{
printf("Essai de deletion d'un element qui n'etait pas deja dans la liste.\n");
exit(-1);
};
  interim = (*liste_elt)->next;
  free (*liste_elt);
  (*liste_elt) = interim;
}


/*
 * Routine d'insertion de p dans la liste pointee
 * par liste_p, dans l'ordre p, a, q.
 */

void croise_liste (liste **liste_p, pt3d *p, pt3d *a, pt3d *q)
{
    liste **liste_elt = liste_p;
    liste *liste_ptr;
    NEW_LISTE(liste_ptr);
    liste_ptr->point = p;

    while ( ((*liste_elt)->point != a) && ((*liste_elt)->point != q))
        liste_elt = &((*liste_elt)->next);

    if ((*liste_elt)->point == a)
    {	/* Le premier est a. */
	if ((*liste_elt)->next->point == q)
	{
	    /* Cas ou q suit betement a: on insere p juste avant a. */
	    liste_ptr->next  = *liste_elt;
	    *liste_elt = liste_ptr;
	}
        else
	{
	    /* Cas ou q est en derniere position, a en premiere.
	     * On integre p juste apres a.
	     */
	    liste_ptr->next    = (*liste_elt)->next;
	    (*liste_elt)->next = liste_ptr;
   	};
    }
    else
    {   /* Le premier est q. */
        if ((*liste_elt)->next->point == a)
	{
	    /* Cas ou a suit betement q: on insere p juste apres a. */
	    liste_ptr->next  = (*liste_elt)->next->next;
	    (*liste_elt)->next->next = liste_ptr;
        }
	else
	{
	    /* Cas ou a est en derniere position, q en premiere.
	     * On integre p juste avant a.
	     */

	    while ((*liste_elt)->point != a)
    	        liste_elt = &((*liste_elt)->next);

	    liste_ptr->next = *liste_elt;
	    *liste_elt = liste_ptr;
        };
    };
}


int TROPLAT(pt3d *a, pt3d *b, pt3d *c)
{
    double aabb = sqrt(DIST_3D(a, b));
    double aacc = sqrt(DIST_3D(a, c));
    double bbcc = sqrt(DIST_3D(b, c));
    return ( 1.1 > MIN((aabb+bbcc)/aacc, 
		       MIN((aabb+aacc)/bbcc, (bbcc+aacc)/aabb)));
}


/* Rend le CARRE de la distance de 2 points dans l'espace */

double dist_3d ( pt3d *a,  pt3d *b)
{
    return (SQR((a)->x-(b)->x) + SQR((a)->y-(b)->y) + SQR((a)->z-(b)->z));
}


/*
 * Rend le voisin commun aux deux sommets, le plus pret
 * des deux sommets, et dans la zone.
 */

unsigned int bon_voisin (maillage *maillage, str_mem *tab, pt3d *ptj, pt3d *voisin)
{
    unsigned int k2 = tab->current;
    unsigned int k3;
    liste *m, *l = maillage->maille[ptj - maillage->first].voisins;
    double temp0 = DBL_MAX;
    double temp1 = DBL_MAX;
    double temp2, temp3;

    do {
        if ((temp3 = DIST_3D (l->point, ptj)) < temp1)
	{
	    m = maillage->maille[voisin - maillage->first].voisins;
	    while ((m != NULL) && (m->point != l->point))
		m = m->next;
				
	    if ( (m != NULL) && ((temp2 = DIST_3D (l->point, voisin)) < temp0) )
	    {
		/* Il faut que ce voisin soit dans le voisinage
		 * pour avoir ses coordonnees planes.
		 */
		k3 = 0;
		while ((k3 < tab->current) && (tab->debut[k3] != m->point)) k3++;
		if (k3 < tab->current)
		{
		    temp0 = temp2;
		    temp1 = temp3;
		    k2 = k3;
		};
	    };
	};
	l = l->next;
    } while (l != NULL);
    return k2;
}


/*
 * Cette fonction verifie qu'il n'y a pas un autre point
 * a l'interieur du futur triangle...
 * et qu'aucun des segments du maillage, 
 * dont les extremites sont dans la zone (on fait ce qu'on peut), 
 * ne coupe l'interieur du triangle.
 */

unsigned int inter_check ( maillage *maillage, str_mem *tab, pt2d *project, 
			   pt3d *pti, pt3d *ptj, pt3d *voisin, 
			   unsigned int i, unsigned int j, unsigned int id_voisin)
{
    unsigned int k3, k = 0;
    unsigned int flag = TRUE;
    liste *pipo;
    
    for (; flag && (k < tab->current); k++)
    {
	flag = !INSIDE( project[i],
		        project[j],
			project[id_voisin],
			project[k]);
        if ( (k != i) && (k != j) && (k != id_voisin) )
	{
	    pipo = maillage->maille[tab->debut[k] - maillage->first].voisins;
	    while ( (flag) && (pipo != NULL) )
	    {
		if ((pipo->point < tab->debut[k]) &&
		    (pipo->point != pti) &&
		    (pipo->point != ptj) &&
		    (pipo->point != voisin))
		{
		    k3 = 0;
		    while ((k3 < tab->current) && (tab->debut[k3] != pipo->point)) k3++;
		    if (k3 < tab->current)
		    {
			if OPPSIDE (project[i], project[j], project[k], project[k3])
			    flag = flag && !OPPSIDE (project[k], project[k3], project[i], project[j]);
			if OPPSIDE (project[id_voisin], project[j], project[k], project[k3])
			    flag = flag && !OPPSIDE (project[k], project[k3], project[id_voisin], project[j]);
		    };
		};
		pipo = pipo->next;
	    };
	};
    };
    return flag;
}
