/*
 * On definit une fonction qui prend en parametre un point
 * de la frontiere, le maillage, et un voisinage de ce point.
 * Elle etend le maillage aux points du voisinage.
 */

unsigned int etend_maillage (	pt3d *point,
				vect *normale, 
				maillage *maillage, 
				str_mem *tab);


/* Routine d'insertion de p dans la liste pointee par liste_elt, entre a et b. */

void insert_liste (liste **liste_p, pt3d *a, pt3d *p, pt3d *b);


/* Routine de remplacement de a par p, dans la liste de b. */

void replace_liste (liste **liste_elt, pt3d *a, pt3d *p);


/* Routine de deletion de a dans la liste de b. */

void delete_liste (liste **liste_elt, pt3d *a);

/*
 * Routine d'insertion de p dans la liste pointee
 * par liste_p, dans l'ordre p, a, q.
 */

void croise_liste (liste **liste_p, pt3d *p, pt3d *a, pt3d *q);


/*
 * Test si 2 points p et q, sont opposes par rapport au segment [a, b].
 */

#define OPPSIDE(a, b, p, q)						\
   ( (((b).x-(a).x)*((p).y-(a).y)-((b).y-(a).y)*((p).x-(a).x)) *	\
     (((b).x-(a).x)*((q).y-(a).y)-((b).y-(a).y)*((q).x-(a).x)) < 0)

/*
 * Test si 2 points p et q, sont du meme cote du segment [a, b].
 */

#define SAMESIDE(a, b, p, q)						\
    ( (((b).x-(a).x)*((p).y-(a).y)-((b).y-(a).y)*((p).x-(a).x)) *	\
      (((b).x-(a).x)*((q).y-(a).y)-((b).y-(a).y)*((q).x-(a).x)) > 0)

/*
 * Test si le point P est a l'interieur du triangle (abc).
 */

#define INSIDE(a, b, c, p)	    \
   ((SAMESIDE((b),(c),(a),(p))) &&  \
    (SAMESIDE((a),(c),(b),(p))) &&  \
    (SAMESIDE((a),(b),(c),(p))))

/*
 * Test si le point P est a l'exterieur du triangle (abc).
 */

#define OUTSIDE(a, b, c, p)	   \
   ((OPPSIDE((b),(c),(a),(p))) ||  \
    (OPPSIDE((a),(c),(b),(p))) ||  \
    (OPPSIDE((a),(b),(c),(p))))

/*
 * Rend le carre du cosinus de l'angle forme par 3 points.
 * (produit scalaire, divise par les normes)
 */

#define COSINUS(a, b, c) \
    ( SQR(((a).x-(b).x)*((c).x-(b).x) + ((a).y-(b).y)*((c).y-(b).y)) / \
      ((SQR((a).x-(b).x) + SQR((a).y-(b).y)) * \
       (SQR((c).x-(b).x) + SQR((c).y-(b).y))) )

#define COSINUS_3D(a, b, c) \
    ( SQR(((a).x-(b).x)*((c).x-(b).x) + \
          ((a).y-(b).y)*((c).y-(b).y) + \
	  ((a).z-(b).z)*((c).z-(b).z)) / \
      ((SQR((a).x-(b).x) + SQR((a).y-(b).y) + SQR((a).z-(b).z)) * \
       (SQR((c).x-(b).x) + SQR((c).y-(b).y) + SQR((c).z-(b).z))) )


/* Je le remplace par quelque chose de plus approprie:
 * Ca doit caracteriser un beau triangle, et etre indep
 * de a, b, et de toute homothetie. Par exemple je cherche
 * a maximiser le plus petit angle, donc je rend le plus
 * petit angle du triangle.
 * En fait le cosinus (au carre) correspondant => ca inverse l'ordre !!!
 */

#define MED_ABS(a, b, c) \
    (MAX(COSINUS (a, b, c), MAX(COSINUS(a, c, b), COSINUS(b, a, c))))


/*
 * Trouve un point qui est sur la frontiere.
 * A partir de begin.
 */

pt3d *pt_front( pt3d *begin, maillage *maillage);


/*
 * Un #define qui calcule la distance au carre entre 2 pt3d*.
 */

/* 
#define DIST_3D(a, b) \
    (SQR((a)->x-(b)->x) + SQR((a)->y-(b)->y) + SQR((a)->z-(b)->z))
*/

#define DIST_3D(a, b) dist_3d ((a), (b))

double dist_3d ( pt3d *a,  pt3d *b);

/* 
 * Dit si un triangle, donne par 3 *pt3d, est trop plat.
 * (il y a des risques selon la normale choisie,
 * que ca fasse tout foirer)
 */

int TROPLAT(pt3d *a, pt3d *b, pt3d *c);

/*
 * Rend le voisin qu'il faut (cf le source).
 */

unsigned int bon_voisin (   maillage *maillage,
			    str_mem *tab,
			    pt3d *ptj,
			    pt3d *voisin);

/*
 * Regarde si un triangle n'est pas coupe.
 */

unsigned int inter_check ( maillage *maillage, str_mem *tab, pt2d *project, 
			   pt3d *pti, pt3d *ptj, pt3d *voisin, 
			   unsigned int i, unsigned int j, unsigned int id_voisin);
			   
