#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "structures.h"
#include "maillage_local.h"


maillage *maillage_init (unsigned int nbpts, pt3d *first)
{
    unsigned int i;
    struct paire *pipolist = (struct paire *) malloc (nbpts * sizeof (struct paire) );
    maillage *pipo = (maillage *) malloc ( sizeof (maillage) );

    if ((pipo == NULL) || (pipolist == NULL))
    {
	printf("Erreur d'allocation memoire dans maillage_init().\n");
	exit(-1);
    };

    for (i=0; i<nbpts; i++) {
	pipolist[i].voisin = NULL;
	pipolist[i].voisins = NULL;
    };
    
    pipo->nb     = nbpts;
    pipo->maille = pipolist;
    pipo->first  = first;
    return pipo;
}


pt2d *projete (	vect *normale, 
		str_mem *tab)
{
    unsigned i;
    double zz;
    vect n1 = {0, 0, 0};
    vect n2;
    pt2d *project = (pt2d *) malloc (tab->current * sizeof(pt2d));

    if (project == NULL)
    {
	printf("Erreur d'allocation memoire dans projete().\n");
	exit(-1);
    };
    /*
     * On calcule dans un premier temps deux vecteurs formant
     * une base orthonormee du plan tangent.
     * 
     * Pour ce faire, on prend le normalise du plus grand des vecteurs:
     * N ^ x,  et N ^ y (produit vectoriel). L'autre vecteur s'obtient
     * facilement par N2 = N ^ N1.
     */

    if (fabs(normale->x) > fabs(normale->y))
    {
	/* On prend n1 = N ^ y */
	zz = normale->z * normale->z + normale->x * normale->x;
	n1.x = - normale->z / zz;
	n1.z = normale->x / zz;
    } else {
	/* On prend n1 = N ^ x */
	zz = normale->z * normale->z + normale->y * normale->y;
	n1.y = normale->z / zz;
	n1.z = - normale->y / zz;
    };

    n2.x = normale->y * n1.z - normale->z * n1.y;
    n2.y = normale->z * n1.x - normale->x * n1.z;
    n2.z = normale->x * n1.y - normale->y * n1.x;

    /* On obtient les coordonnees planes par des produits scalaires
     * avec N1 et N2, de OPi, je stocke ces coordonnes (temporaires)
     * dans un tableau.
     */

    for (i=0; i<tab->current; i++)
    {
	project[i].x =	tab->debut[i]->x * n1.x +
			tab->debut[i]->y * n1.y +
			tab->debut[i]->z * n1.z;
	project[i].y =	tab->debut[i]->x * n2.x +
			tab->debut[i]->y * n2.y +
			tab->debut[i]->z * n2.z;
    };

    return project;
}


void maillage_local (	vect *normale, 
			str_mem *tab,
			maillage *pre_maillage)
{
    int i, j, k, l;

    liste **liste_elt;
    liste *prochain, *prochain2;
    unsigned int frontiere1, frontiere2;
    pt2d *project   	    = projete( normale, tab);
    unsigned int *tri	    = (unsigned int *) 
	    malloc (tab->current * sizeof(unsigned int));
    unsigned int *maillage  = (unsigned int *) 
	    malloc (tab->current * tab->current * sizeof(unsigned int));

    if ( (tri == NULL) || (maillage == NULL) )
    {
	printf("Erreur d'allocation memoire dans maillage_local().\n");
	exit(-1);
    };

    /*
     * Ensuite, on trie les points selon x.
     * (Pour effectuer une triangulation par balayage)
     * 
     * NB1: On prie tres fort pour qu'il n'y ait pas 2 points de meme
     *      abscisse, sinon ca plante (pas au tri, plus tard) :-)
     * 
     * NB2: Dans un premier temps, je fais un bete tri en n^2...
     *      Ca pourra etre ameliore par la suite.
     */

    for (i=0; i<tab->current; i++)
    {
	j=0;
	while ( (j<i) && (project[i].x < project[tri[j]].x) )
	    j++;
	
	for (k=i; k>j; k--)
	    tri[k] = tri[k-1];

	tri[j] = i;
    };

    
    /*
     * On initialise le maillage.
     * 
     * NB: Il faut au minimum 3 points !!!!!!!!!!!!!!!
     * (faire le test avant d'appeler la fonction)
     * 
     * Pour chaque point, je stocke d'abord le nombre courant
     * de ses voisins, puis leurs indices, en parcourant dans
     * le sens inverse des aiguilles d'une montre, en
     * commencant a midi.
     */

    for (i=0; i<tab->current; i++)
	/*
	 * On pourrait se passer de ca => lineaire a la place de n^2.
	 */
	for (j=0; j<tab->current; j++)
	    maillage[i*tab->current+j] = 0;

    maillage[0*tab->current] = 2;
    maillage[1*tab->current] = 2;
    maillage[2*tab->current] = 2;

    /* Deux types de triangles sont possibles */

    if ( (project[tri[1]].y - project[tri[0]].y) *
         (project[tri[2]].x - project[tri[0]].x) <
	 (project[tri[2]].y - project[tri[0]].y) *
	 (project[tri[1]].x - project[tri[0]].x) )
    {
	maillage[0*tab->current+1] = 1;
	maillage[0*tab->current+2] = 2;
	maillage[1*tab->current+1] = 0;
	maillage[1*tab->current+2] = 2;
	maillage[2*tab->current+1] = 0;
	maillage[2*tab->current+2] = 1;
    }
    else 
    {
	maillage[0*tab->current+1] = 2;
	maillage[0*tab->current+2] = 1;
	maillage[1*tab->current+1] = 0;
	maillage[1*tab->current+2] = 2;
	maillage[2*tab->current+1] = 1;
	maillage[2*tab->current+2] = 0;
    };

    /*
     * Boucle d'insertion des points dans la structure (balayage).
     */

    for (i=3; i<tab->current; i++)
    {
	/*
	 * On cherche le point le plus "haut" (vu de i) sur l'enveloppe
	 * convexe des precedents, et apres, on descend.
	 */

	k = i-1;
	j = maillage[k*tab->current + 1];

	while ( (k != 0) &&
	      ( (project[tri[i]].y - project[tri[k]].y) *
		(project[tri[i]].x - project[tri[j]].x) >
		(project[tri[i]].y - project[tri[j]].y) *
		(project[tri[i]].x - project[tri[k]].x) ) )
	{
	    k = j;
	    j = maillage[k*tab->current + 1];
	};

	/*
	 * k pointe desormais vers le point le plus "haut"
	 * (vu de i, le point a inserer)
	 * (mais qui reste sur l'enveloppe convexe).
	 * j, celui juste au dessus.
	 * On fait la descente, en actualisant la structure.
	 * 
	 * NB: Dans un premier temps, je ne fais pas d'optimisation
	 *     locale de la triangulation, mais ca pourra
	 *     se faire (si besoin).
	 */

	/* J'entame la descente, jusqu'a i-1. */

	do {
	    maillage[i*tab->current + (++maillage[i*tab->current])] = k;
	    maillage[k*tab->current + (++maillage[k*tab->current])] = i;
	    j = k;
	    k = maillage[k*tab->current + maillage[k*tab->current] - 1];
	} while (j != i-1);

	/* Je continue la descente... c'est plus delicat. */

	while ( (k != -1) &&
	      ( (project[tri[i]].y - project[tri[k]].y) *
		(project[tri[i]].x - project[tri[j]].x) >
		(project[tri[i]].y - project[tri[j]].y) *
		(project[tri[i]].x - project[tri[k]].x) ) )
	{
    	    /* Je relie k <-> i. */

	    maillage[i*tab->current + (++maillage[i*tab->current])] = k;

	    /* J'insere i, dans la liste de k. */

	    l = ++maillage[k*tab->current];
	    do {
		maillage[k*tab->current + l] = maillage[k*tab->current + l - 1];
		--l;
	    } while (j != maillage[k*tab->current + l]);
	    
	    maillage[k*tab->current + l] = i;

	    /* Je passe au couple suivant. */

	    j = k;
	    if ((l == 1) && (maillage[k*tab->current + l] == 0))
		k = -1; /* flag pour s'arreter */
	    else
		k = maillage[k*tab->current + l - 1];
	};
    };


    /*
     * Le maillage local est fini.
     * Il n'est pas optimal...
     * 
     * Je renvoie le resultat, sous la forme qu'il faut.
     */

    free (project);

    for (i=0; i<tab->current; i++)
    {
	liste_elt = &((pre_maillage->maille)[tab->debut[tri[i]] - pre_maillage->first].voisins);
	for (j=1; j<=maillage[i*tab->current]; j++)
	{
	    NEW_LISTE(prochain);
	    prochain->point = tab->debut[tri[maillage[i*tab->current + j]]];
	    *liste_elt = prochain;
	    liste_elt = &(prochain->next);
/* Ancien code
NEW_LISTE(*liste_elt);
(*liste_elt)->point = tab->debut[tri[maillage[i*tab->current + j]]];
liste_elt = &((*liste_elt)->next);
*/
	};
	*liste_elt = NULL;
    };

    /*
     * On teste ceux qui sont sur la frontiere.
     * On part du point d'x minimum, et on longe la frontiere, 
     * jusqu'a retomber dessus.
     */

    frontiere1 = tab->debut[tri[0]] - pre_maillage->first;
    frontiere2 = pre_maillage->maille[frontiere1].voisins->point - pre_maillage->first;
    pre_maillage->maille[frontiere1].voisin = frontiere2 + pre_maillage->first;
    do {
	/*
	 * Trouver le point voisin de frontiere2, != frontiere1, 
	 * et qui ne soit pas relie a frontiere1.
	 * mais consecutif de frontiere1 dans la liste des voisins de fr2.
	 */
	 
	prochain = pre_maillage->maille[frontiere2].voisins;
	while ( (prochain->next != NULL) &&
		(prochain->next->point != frontiere1 + pre_maillage->first) )
	    prochain = prochain->next;

	/* Je teste si prochain est relie a frontiere1, et si non, je le
	 * garde, et si oui, je prends prochain->next->next (modulo...).
	 */

	prochain2 = pre_maillage->maille[prochain->point - pre_maillage->first].voisins;
	i=0; /* flag */
	while ((i==0) && (prochain2 != NULL))
	{
	    if (prochain2->point == (frontiere1 + pre_maillage->first))
		i=1;
	    prochain2 = prochain2->next;
	};

	if (i==1)
	    if (prochain->next == NULL) {
		    prochain = pre_maillage->maille[frontiere2].voisins->next;}
	    else {
	      if (prochain->next->next != NULL) {
		prochain = prochain->next->next;}
	      else prochain = pre_maillage->maille[frontiere2].voisins;
	    };

	pre_maillage->maille[frontiere2].voisin = prochain->point;
	frontiere1 = frontiere2;
	frontiere2 = prochain->point - pre_maillage->first;
    } while (frontiere2 != tab->debut[tri[0]] - pre_maillage->first);
    
    free (tri);
    free (maillage);
}

void maillage_triangle (    vect *normale, 
			    str_mem *tab,
			    maillage *pre_maillage)
{
    int i, j, k;
    liste *prochain, *prochain2;
    pt2d *project   	    = projete( normale, tab);
    unsigned int *tri	    = (unsigned int *) 
	    malloc (tab->current * sizeof(unsigned int));

    if (tri == NULL)
    {
	printf("Erreur d'allocation memoire dans maillage_triangle().\n");
	exit(-1);
    };

    /*
     * Ensuite, on trie les points selon x.
     * Ca garantie que le triangle est vide.
     * On pourrait plus simplement prendre les 3 plus petits...
     * => lineaire :-)
     */

    for (i=0; i<tab->current; i++)
    {
	j=0;
	while ( (j<i) && (project[i].x < project[tri[j]].x) )
	    j++;
	
	for (k=i; k>j; k--)
	    tri[k] = tri[k-1];

	tri[j] = i;
    };

    pre_maillage->maille[tab->debut[tri[0]]-pre_maillage->first].voisin = tab->debut[tri[1]];
    pre_maillage->maille[tab->debut[tri[1]]-pre_maillage->first].voisin = tab->debut[tri[2]];
    pre_maillage->maille[tab->debut[tri[2]]-pre_maillage->first].voisin = tab->debut[tri[0]];

    NEW_LISTE(prochain);
    NEW_LISTE(prochain2);
    prochain->point = tab->debut[tri[1]];
    prochain->next = prochain2;
    prochain2->point = tab->debut[tri[2]];
    prochain2->next = NULL;
    pre_maillage->maille[tab->debut[tri[0]]-pre_maillage->first].voisins = prochain;

    NEW_LISTE(prochain);
    NEW_LISTE(prochain2);
    prochain->point = tab->debut[tri[0]];
    prochain->next = prochain2;
    prochain2->point = tab->debut[tri[2]];
    prochain2->next = NULL;
    pre_maillage->maille[tab->debut[tri[1]]-pre_maillage->first].voisins = prochain;

    NEW_LISTE(prochain);
    NEW_LISTE(prochain2);
    prochain->point = tab->debut[tri[0]];
    prochain->next = prochain2;
    prochain2->point = tab->debut[tri[1]];
    prochain2->next = NULL;
    pre_maillage->maille[tab->debut[tri[2]]-pre_maillage->first].voisins = prochain;

    free (project);
    free (tri);
}

