/*
 * Generation d'ensembles de points (surfaces classiques).
 * Tests de l'algo sur ces points.
 * Affichage.
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "structures.h"
#include "vect_normal.h"
#include "genere.h"
#include "struct_local.h"
#include "xdebug.h"
#include "maillage_local.h"
#include "etend_maillage.h"
#include "main.h"
#include "optimise.h"
#include "topo.h"

int compteur_arret, compt_old;
unsigned int opti_arret = 1000000;

/* Pour le debug dans etend_maillage.c */
maillage *maillage_demo;


/* On arrete d'etendre au bout de compteur_arret segments,
   et on arrete les optimisations au bout de opti_arret swaps. */


int main (int argc, char **argv)
{
  int i;
  unsigned short fait_kekchoz;
  unsigned int nombre_points;
  double distance;
  double distmax;
  pt3d *para, *essai;
  str_mem neighbours;
  pt_dist pt_essai, *englobe;
  strloc *arbre;
  vect *normale;
  /*  maillage *maillage_demo; */
  pt3d **paratab;

  if (argc != 6) {
    printf("Usage: maillage arg1 arg2 arg3 arg4 arg5\n");
    printf("  arg1 = distance initiale [1]\n");
    printf("  arg2 = distance finale   [10]\n");
    printf("  arg3 = nombre de points  [500]\n");
    printf("  arg4 = proportionnel au nombre maximum\n");
    printf("         de points a mailler [-1]\n");
    printf("  arg5 = type d'entree:\n");
    printf("         0 -> stdin\n");
    printf("         1 -> demi-sphere\n");
    printf("         2 -> paraboloide\n");
    printf("         3 -> coin de cube\n");
    printf("         4 -> cube complet\n");
    exit(-1);
  };
  
  distance = atof(argv[1]);
  distmax  = atof(argv[2]);
  nombre_points  = atoi(argv[3]);
  compteur_arret = atoi(argv[4]);

  switch (atoi(argv[5])) {
  case 0: para = gen_file (&nombre_points, stdin); break;
  case 1: para = gen_sph (nombre_points, 0,0,0,10); break;
  case 2: para = gen_para (nombre_points, -10, 10, -10, 10, 0.1, 0.1); break;
  case 3: para = gen_cube (nombre_points, 10); break;
  case 4: para = gen_fullcube (nombre_points, 10); break;
  default: para = gen_file (&nombre_points, stdin);
  };


printf("Nombre de points a traiter: %d\n", nombre_points);

  paratab = (pt3d **) (malloc (nombre_points * sizeof (pt3d *)));
  
  if (paratab == NULL) {
    printf("Erreur d'allocation memoire dans main().\n");
    exit(-1);
  };

printf("Calcul de la structure de localisation en cours...");
  
  for (i=0; i<nombre_points; i++)
    paratab[i] = para + i;
  
  englobe = mini_cube (paratab, nombre_points);
  arbre = calcul_struct (paratab, nombre_points, englobe);

printf(" fini.\n");
  
  pt_essai.pt.x = para[0].x;
  pt_essai.pt.y = para[0].y;
  pt_essai.pt.z = para[0].z;
  pt_essai.d    = distance / 2;
  
/* On initialise le maillage. */

  init_strmem (&neighbours);
  voisins ( &pt_essai, &neighbours, englobe, arbre);
  normale = mat ( para, &neighbours);
  cone_elimine (&pt_essai.pt, normale, &neighbours);
  
  /* 3 DEVRAIT suffir, normalement */
  
  while (neighbours.current < 4)
    {
      free (normale);
      free_strmem (&neighbours);
      init_strmem (&neighbours);
      pt_essai.d *= 1.1;
      voisins ( &pt_essai, &neighbours, englobe, arbre);
      normale = mat ( para, &neighbours);
      cone_elimine ( &pt_essai.pt, normale, &neighbours);
    };

    free (normale);
    normale = mat ( para, &neighbours);
    
  maillage_demo = maillage_init ( nombre_points, para);

  maillage_local ( normale, &neighbours, maillage_demo);
  /*
  maillage_triangle ( normale, &neighbours, maillage_demo);
  */

  free_strmem (&neighbours);
  free (normale);
  essai = maillage_demo->first;

  fait_kekchoz = TRUE;
  
  /* On le poursuit, tant qu'on peut. */
  
  while (distance < distmax)   /* limite arbitraire !!!! */ 
    {
      printf ("distance courante == %.20f\n", distance);
      while ( (0 < compteur_arret) && 
	      ( ((NULL != (essai = pt_front (essai, maillage_demo))) ||
		 (fait_kekchoz) )))
	{
	  if (essai == NULL)
	    {
	      fait_kekchoz = FALSE;
	      essai = pt_front (maillage_demo->first, maillage_demo);
	    };
	  if (essai != NULL)
	    {
	      init_strmem (&neighbours);
	      pt_essai.pt.x = essai->x;
	      pt_essai.pt.y = essai->y;
	      pt_essai.pt.z = essai->z;
	      pt_essai.d    = distance / 2;
	      
	      if( (compteur_arret/100) != (compt_old/100))
		{
		  printf("Reste a mailler: %d\n", compteur_arret);
		  compt_old = compteur_arret;
		};
	      voisins ( &pt_essai, &neighbours, englobe, arbre);
	      
	      if (neighbours.current > 4)
		{
		  normale = mat (essai,&neighbours);
		  /*		  cone_elimine (&pt_essai.pt, normale, &neighbours);*/
		  if (etend_maillage(essai, normale, maillage_demo, &neighbours) != 0)
		    {
		      /*
		       * On repart des le debut. Comme ca, si c'est nul, on continue,
		       * jusqu'a ce qu'il n'y ai plus aucun point de la frontiere, 
		       * qui ne conduise a des nouveaux segments mailles.
		       */
		      fait_kekchoz = TRUE; ++essai;
		      /*
			essai = maillage_demo->first;
			free(normale);
			*/
		    }
		  else ++essai;
		  free(normale);
		}
	      else ++essai;
	      free_strmem ( &neighbours);
	    };
	};
      
      /*            
		    printf("Phase d'optimisation.\n");
		    triangle_front (maillage_demo);
		    optimise (maillage_demo);
		    cut_front (maillage_demo, 0.9); 
		    bouche_front (maillage_demo, 0.9);
		    */
	
      essai = maillage_demo->first;
      distance *= 1.25;
    };

  /*  
    printf ("Phase d'optimisation finale.\n");
    for (i=0;i<3;i++)
	optimise (maillage_demo);

    for (i=0;i<1;i++)
	cut_front (maillage_demo, 0.9);

    for (i=0;i<1;i++)
	bouche_front (maillage_demo, 0.9);
	*/
      
  printf ("Calcul fini.\n");
  
  free_strloc (arbre);
  free (arbre);
  
  x_dessine (maillage_demo);

/*
 * Liberer le maillage => y'a pas la fonction pour !!!
 */

  return 0;
}
