#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "structures.h"
#include "struct_local.h"
#include "optimise.h"
#include "vect_normal.h"
#include "etend_maillage.h"
#include "maillage_local.h"

extern opti_arret;

void optimise (maillage *maillage)
{
  unsigned i;
  double interim;
  pt3d *p[4];
  liste *pointeur, *pipo;
  pt2d *project;
  vect normale; 
  str_mem courant;
  
  courant.debut = &p[0];
  courant.max = 4;
  courant.current = 4;
  
  for (i=0; i<maillage->nb; i++)
    if ( ((pointeur = maillage->maille[i].voisins) != NULL) &&
	 (pointeur->next != NULL) &&
	 (pointeur->next->next != NULL) )
      {
	p[0] = maillage->first + i;
	while (pointeur != NULL)
	  {
	    p[1] = pointeur->point;
	    if (pointeur->next != NULL) {
	      p[2] = pointeur->next->point;
	      if (pointeur->next->next != NULL) {
		p[3] = pointeur->next->next->point;
	      } else
		p[3] = maillage->maille[i].voisins->point;
	    } else {
	      p[2] = maillage->maille[i].voisins->point;
	      p[3] = maillage->maille[i].voisins->next->point;
	    };
	    
	    /* A noter, que p[0]<p[2] devrait diminuer
	     * les passages "inutiles", mais bon ca se discute.
	     */
	    
	    /* A tout hasard, je verifie que p1 n'est pas deja relie a p3, on sait jamais. */
	    
	    pipo = maillage->maille[p[1]-maillage->first].voisins;
	    while ((pipo != NULL) && (pipo->point != p[3]))
	      pipo = pipo->next;
	    
	    /* Si p[0] est sur la frontiere, faut simplement faire gaffe
	     * a ce que p[2] n'y soit pas non plus, consecutivement.
	     */
	    if ((pipo == NULL) &&
		(maillage->maille[i].voisin != p[2]) &&
		(maillage->maille[p[2]-maillage->first].voisin != p[0]))
	      {
		/* On verifie que p2, projete sur le plan (p0,p1,p3),
		 * est bien dans la bonne zone.
		 */
		/* Pour la normale, on prend p0p1 ^ p0p3, normalise. */
		
		normale.x = (p[1]->y - p[0]->y) * (p[3]->z - p[0]->z) -
		  (p[1]->z - p[0]->z) * (p[3]->y - p[0]->y);
		normale.y = (p[1]->z - p[0]->z) * (p[3]->x - p[0]->x) -
		  (p[1]->x - p[0]->x) * (p[3]->z - p[0]->z);
		normale.z = (p[1]->x - p[0]->x) * (p[3]->y - p[0]->y) -
		  (p[1]->y - p[0]->y) * (p[3]->x - p[0]->x);
		
		interim = sqrt(SQR(normale.x) + SQR(normale.y) + SQR(normale.z));
		normale.x /= interim;
		normale.y /= interim;
		normale.z /= interim;
		
		project = projete (&normale, &courant);
		
		/* Pour la normale, on prend le plan moyen des 4 points. */
		
		/* Cette methode a l'air de bugguer des fois... HUM !
		   normale = mat (p[0], &courant);
		   project = projete (normale, &courant);
		   free(normale);
		   */
		
		if ( (OPPSIDE (project[1], project[3], project[0], project[2])) &&
		     (OPPSIDE (project[0], project[2], project[1], project[3])) &&
		(SAMESIDE(project[0], project[3], project[1], project[2])) &&
		     (SAMESIDE(project[0], project[1], project[3], project[2])) &&
		(MAX(MED_ABS(project[0], project[1], project[3]),
		     MED_ABS(project[2], project[1], project[3])) <
		 MAX(MED_ABS(project[0], project[1], project[2]),
		     MED_ABS(project[0], project[3], project[2]))) &&
		     
		     /*			 (DIST_3D(p[0], p[2]) > DIST_3D(p[1], p[3])) && */
		     (opti_arret > 0))
		  { /* On swap */
		    --opti_arret;
		    delete_liste (&(maillage->maille[i].voisins), p[2]);
		    delete_liste (&(maillage->maille[p[2]-maillage->first].voisins), p[0]);
		    insert_liste (&(maillage->maille[p[1]-maillage->first].voisins), p[2], p[3], p[0]);
		    insert_liste (&(maillage->maille[p[3]-maillage->first].voisins), p[0], p[1], p[2]);
		  }
		else
		  pointeur = pointeur->next;
		
		free(project);
		/*		free_strmem (&courant); */
	      }
	    else
	      pointeur = pointeur->next;
	  };
      };
}


void cut_front (maillage *maillage, double percent)
{
    unsigned int i;
    pt3d *voisin;
    liste *commun, *interim, *communopt;
    double temp0, temp1;
    
    for (i=0; i<maillage->nb; i++)
	if ((voisin = maillage->maille[i].voisin) != NULL)
	{
	    commun  = maillage->maille[i].voisins;
	    interim = maillage->maille[voisin-maillage->first].voisins;
	    communopt = NULL;
	    temp0 = temp1 = DBL_MAX;

	    while (interim != NULL)
	    {
		if ( (interim->point == commun->point) &&
		     (maillage->maille[commun->point-maillage->first].voisin == NULL) &&
		     (DIST_3D (interim->point, voisin) < temp0) &&
		     (DIST_3D (interim->point, maillage->first + i) < temp1) )
		{
		    temp0 = DIST_3D (interim->point, voisin);
		    temp1 = DIST_3D (interim->point, maillage->first + i);
		    communopt = commun;
		};
		commun = commun->next;
		if (commun == NULL)
		{
		    interim = interim->next;
		    commun = maillage->maille[i].voisins;
		};
	    };

	    if ((communopt != NULL) &&
		((sqrt(DIST_3D(i+maillage->first, voisin)) >
		  percent * ( sqrt( DIST_3D (i+maillage->first, communopt->point)) +
			      sqrt( DIST_3D (communopt->point, voisin))) )))
	      {
		maillage->maille[i].voisin = communopt->point;
		maillage->maille[communopt->point-maillage->first].voisin = voisin;
		delete_liste ( &maillage->maille[i].voisins, voisin);
		delete_liste ( &maillage->maille[voisin-maillage->first].voisins,
			       i+maillage->first);
	      };
	};
}


void bouche_front (maillage *maillage, double percent)
{
    unsigned int i;
    pt3d *voisin, *voisin2;
    liste *interim;
    
    for (i=0; i<maillage->nb; i++)
	if ((voisin = maillage->maille[i].voisin) != NULL)
	{
	    voisin2 = maillage->maille[voisin-maillage->first].voisin;
	    
	    interim = maillage->maille[i].voisins;
	    while ((interim != NULL) && (interim->point != voisin2))
		interim = interim->next;

	    if ((interim == NULL) &&
		(COSINUS_3D(*(i+maillage->first), *voisin, *voisin2) > percent) &&
		(DIST_3D(i+maillage->first, voisin2) < DIST_3D(i+maillage->first, voisin)))
/*		(DIST_3D(i+maillage->first, voisin2) <
		 percent * MAX(	DIST_3D(i+maillage->first, voisin), 
				DIST_3D(voisin, voisin2) ) ))
*/	    {
		maillage->maille[i].voisin = voisin2;
		maillage->maille[voisin-maillage->first].voisin = NULL;

		interim = maillage->maille[i].voisins;
		while ( maillage->maille[interim->point-maillage->first].voisin != 
			(i + maillage->first))
		    interim = interim->next;

		insert_liste( &maillage->maille[i].voisins, voisin, voisin2, interim->point);
		insert_liste( &maillage->maille[voisin2-maillage->first].voisins,
			      voisin, i+maillage->first,
			      maillage->maille[voisin2-maillage->first].voisin);
		if ( maillage->maille[maillage->maille[maillage->maille[i].voisin-maillage->first].voisin-maillage->first].voisin == i+maillage->first)
		{
		    maillage->maille[maillage->maille[maillage->maille[i].voisin-maillage->first].voisin-maillage->first].voisin = NULL;
		    maillage->maille[maillage->maille[i].voisin-maillage->first].voisin = NULL;
		    maillage->maille[i].voisin = NULL;
		};
	    };
	};
}


void triangle_front (maillage *maillage)
{
  unsigned int i;

  for ( i=0; i<maillage->nb; i++)
    {
      if ( (maillage->maille[i].voisin != NULL) &&
	   (maillage->maille[maillage->maille[maillage->maille[i].voisin-maillage->first].voisin-maillage->first].voisin-maillage->first == i) )
	{
printf("Triangle elimine sur la frontiere... : ca ne devrait jamais arriver! \n");
	  maillage->maille[maillage->maille[maillage->maille[i].voisin-maillage->first].voisin-maillage->first].voisin = NULL;
	  maillage->maille[maillage->maille[i].voisin-maillage->first].voisin = NULL;
	  maillage->maille[i].voisin = NULL;
	};
    };
}
