/*
 * Ce fichier contient les fonctions utiles pour manipuler la structure
 * de localisation des points.
 * 
 * Plutot que d'utiliser les fonctions de gestion memoire du systeme,
 * la gerer soi-meme => plus speed (mais plus bug_risky).
 * 
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "structures.h"
#include "struct_local.h"


/*
 * A chaque cube de la structure sont associees les coordonnees
 * du centre, ainsi que la largeur d'un cote du cube.
 * 
 * Ces coordonnees sont recalculees au vol (pas maintenues dans
 * la structure).
 * 
 */


void free_strloc (strloc *arbre)
{
    if ((arbre->flag) == 0)	/* cube */
    {
	free_strloc ( arbre->kwa.cube );
	free_strloc ( arbre->kwa.cube + 1 );
	free_strloc ( arbre->kwa.cube + 2 );
	free_strloc ( arbre->kwa.cube + 3 );
	free_strloc ( arbre->kwa.cube + 4 );
	free_strloc ( arbre->kwa.cube + 5 );
	free_strloc ( arbre->kwa.cube + 6 );
	free_strloc ( arbre->kwa.cube + 7 );
	free	    ( arbre->kwa.cube );
    };
/*
 *    free (arbre);
 * Ne pas faire ca, car ce sont des tableaux.
 * Donc pour liberer la racine, il faut le faire separemment.
 * 
 */ 

}


/*
 * La fonction init_strloc() rend un pointeur
 * vers un arbre initialise.
 * 
 */

strloc *init_strloc (void)
{
    strloc  *init = (strloc *) (malloc (sizeof (strloc)));

    if (init != NULL) {
	init->flag = 1;
	init->kwa.feuille = NULL;
    }
    else {
	printf("Erreur d'allocation memoire dans init_strloc().\n");
	exit(-1);
    };

    return init;
}


/*
 * La fonction init_strloc_8() rend un pointeur
 * vers un tableau de 8 strloc initialises.
 * 
 */

strloc *init_strloc_8 (void)
{
    strloc *init = (strloc *) (malloc ( 8 * sizeof (strloc) ));

    if (init != NULL) {
	unsigned short i;
	for (i=0; i<8; i++) {
	    init[i].flag = 1;
	    init[i].kwa.feuille = NULL;
	};
    }
    else {
	printf("Erreur d'allocation memoire dans init_strloc_8().\n");
	exit(-1);
    };

    return init;
}


/*
 * Cette fonction insere un point dans la structure.
 * 
 * ATTENTION, aucun test n'est fait pour verifier que le point est bien
 * dans le cube de depart. Dans le cas contraire => descente infinie
 * dans l'arbre...
 * 
 */

void insert_pt (pt3d	*point,
		pt_dist *cube, 
		strloc	*arbre)
{
    switch (arbre->flag) {
	case 0:
	    {
		unsigned short indice = 0;
		pt_dist new_cube;

		new_cube.d    = cube->d / 2;
		new_cube.pt.x = cube->pt.x - new_cube.d / 2;
		new_cube.pt.y = cube->pt.y - new_cube.d / 2;
		new_cube.pt.z = cube->pt.z - new_cube.d / 2;

		if (point->x > cube->pt.x) {
		     indice += 4;
		     new_cube.pt.x += new_cube.d;
		};

		if (point->y > cube->pt.y) {
		    indice += 2;
		    new_cube.pt.y += new_cube.d;
		};

		if (point->z > cube->pt.z) {
		    indice += 1;
		    new_cube.pt.z += new_cube.d;
		};

		insert_pt(  point,
			    &new_cube,
			    arbre->kwa.cube + indice);
	    };
	    break;

	case 1:
	    if ((arbre->kwa.feuille) == NULL)
	    {
		/*
		 * Feuille vide.
		 */

		arbre->kwa.feuille = point;
	    }
	    else
	    {
		/*
		 * La feuille est occupee, il faut approfondir.
		 * Sauf si les deux points sont	egaux.
		 */

	      pt3d *pt1 = arbre->kwa.feuille;

	      if ( (pt1->x != point->x) || 
		   (pt1->y != point->y) ||
		   (pt1->z != point->z) )
		{
		arbre->flag = 0;
		arbre->kwa.cube = init_strloc_8 ();
		insert_pt( pt1,	  cube, arbre);
		insert_pt( point, cube, arbre);
		};
	    };
	    break;

	default:
	    printf("Y'a clairement une erreur... insert_pt()\n");
	    break;
    };
}



/*
 * A partir d'un pt_dist( un pt3d* "p" et un double "d"), fournit
 * les points a distance <d de p et dans le cube "cube" (de centre
 * cube->pt, et de cote cube->d).
 *
 * On n'utilise pas la metrique euclidienne,  mais L_infini:
 * d(p, q) = max(|p.x-q.x|, |p.y-q.y|, |p.z-q.z|)
 * 
 */
 
 
void voisins (	pt_dist *point,
		str_mem *tab,
		pt_dist *cube,
		strloc	*arbre)
{
    switch (arbre->flag) {
	case 0:
	    {
		unsigned char testy, testY, testz, testZ;
		double cote	    = cube->d / 2;
		double demi_cote    = cote / 2;
		double limite	    = point->d + demi_cote;
		pt_dist new_cube;

		new_cube.pt.x = cube->pt.x - demi_cote;
		new_cube.pt.y = cube->pt.y - demi_cote;
		new_cube.pt.z = cube->pt.z - demi_cote;
		new_cube.d    = cote;

		/*
		 * Il faut parcourir les sous-cubes.
		 */

		testy = fabs(new_cube.pt.y - point->pt.y) < limite;
		testY = fabs(new_cube.pt.y - point->pt.y + cote) < limite;
		testz = fabs(new_cube.pt.z - point->pt.z) < limite;
		testZ = fabs(new_cube.pt.z - point->pt.z + cote) < limite;


		if (fabs(new_cube.pt.x - point->pt.x) < limite)
		{
		    if (testy)
		    {
			if (testz)
			    voisins (	point,
					tab,
					&new_cube, 
					arbre->kwa.cube);

			if (testZ)
			{
			    new_cube.pt.z += cote;
			    voisins (	point,
					tab,
					&new_cube, 
					arbre->kwa.cube + 1);
			    new_cube.pt.z -= cote;
			};
		    };
		    
		    if (testY)
		    {
			new_cube.pt.y += cote;
			if (testz)
			    voisins (	point,
					tab,
					&new_cube,
					arbre->kwa.cube + 2);

			if (testZ)
			{
			    new_cube.pt.z += cote;
			    voisins (	point,
					tab,
					&new_cube,
					arbre->kwa.cube + 3);
			    new_cube.pt.z -= cote;
			};
    			new_cube.pt.y -= cote;
		    };

		};

		new_cube.pt.x += cote;

		if (fabs(new_cube.pt.x - point->pt.x) < limite)
		{
		    if (testy)
		    {
			if (testz)
			    voisins (	point,
					tab,
					&new_cube,
					arbre->kwa.cube + 4);

			if (testZ)
			{
			    new_cube.pt.z += cote;
			    voisins (	point,
					tab,
					&new_cube,
					arbre->kwa.cube + 5);
			    new_cube.pt.z -= cote;
			};
		    };
		    
		    if (testY)
		    {
			new_cube.pt.y += cote;
			if (testz)
			    voisins (	point,
					tab,
					&new_cube, 
					arbre->kwa.cube + 6);
			if (testZ)
			{
			    new_cube.pt.z += cote;
			    voisins (	point,
					tab,
					&new_cube, 
					arbre->kwa.cube + 7);
			};
		    };
		};
	    };
	    break;

	case 1:
	    if ((arbre->kwa.feuille) != NULL)
	    {
		/*
		 * Feuille non vide.
		 * 
		 */

		double limite = point->d;

		if ((fabs(arbre->kwa.feuille->x - point->pt.x) < limite) &&
		    (fabs(arbre->kwa.feuille->y - point->pt.y) < limite) &&
		    (fabs(arbre->kwa.feuille->z - point->pt.z) < limite))
		    {
			/*
			 * Realloue de la memoire, si besoin.
			 */
			if (tab->max <= tab->current)
			    if (NULL == (
			       tab->debut = realloc (
						(void *) tab->debut,
			    (tab->max += STR_MEM_STEP) * sizeof (pt3d *)
						    )
					)
			    ) {
				printf(
    		    "Erreur d'allocation memoire dans voisins().\n");
				exit(-1);
			    };
			
			tab->debut[tab->current++] = arbre->kwa.feuille;
		    };
	    };
	    break;

	default:
	    printf("Y'a clairement une erreur...\n");
	    break;
    };
   
}


		 
void free_strmem (str_mem *tab)
{
    free ( tab->debut );
}


/*
 * La fonction init_strmem() reserve un tableau initial.
 * 
 */

void init_strmem (str_mem *tab)
{
    if (NULL == (
       tab->debut = (pt3d **) calloc (STR_MEM_STEP, sizeof (pt3d *)) ))
    {
	printf("Erreur d'allocation memoire dans init_strmem.\n");
	exit(-1);
    };
    tab->current = 0;
    tab->max = STR_MEM_STEP;
}


/*
 * Pour calculer effectivement la structure.
 * 
 */
 
strloc *calcul_struct (	pt3d **pts, 
			unsigned int nbpts, 
			pt_dist *gros_cube)
{

    /*
     * 
     *      Voir s'il faut remplacer init_strloc_8() par init_strloc().
     * 
     * 
     */

    strloc *sommet = init_strloc_8 ();
    unsigned int i;

    for (i=0; i<nbpts; i++)
	insert_pt (pts[i], gros_cube, sommet);
    return sommet;
}
 
 /*
  * Rend un cube minimum (centre et cote) qui englobe les points.
  * 
  */
  
pt_dist *mini_cube (	pt3d **pts, 
			unsigned int nbpts)
{
    unsigned int i;
    double xmin = pts[0]->x;
    double xmax = pts[0]->x;
    double ymin = pts[0]->y;
    double ymax = pts[0]->y;
    double zmin = pts[0]->z;
    double zmax = pts[0]->z;
    pt_dist *cube = (pt_dist *) (malloc (sizeof (pt_dist)));

    if (cube == NULL)
    {
	printf("Erreur d'allocation memoire dans mini_cube().\n");
	exit(-1);
    };
   
    for (i=1; i<nbpts; i++)
    {
	xmin = MIN (xmin, pts[i]->x);	xmax = MAX (xmax, pts[i]->x);
	ymin = MIN (ymin, pts[i]->y);	ymax = MAX (ymax, pts[i]->y);
	zmin = MIN (zmin, pts[i]->z);	zmax = MAX (zmax, pts[i]->z);
    };

    cube->d    = MAX( MAX(xmax-xmin, ymax-ymin), zmax-zmin);
    cube->pt.x = (xmax + xmin) / 2;
    cube->pt.y = (ymax + ymin) / 2;
    cube->pt.z = (zmax + zmin) / 2;

    return cube;
}


/*
 * La fonction suivante permet de proposer une distance
 * caracteristique de l'ensemble de points, a savoir:
 * a partir d'un ensemble de points echantillonnes
 * (suffisamment grand, mais pas trop, ie: la racine
 * carree du nombre de points qu'on a, mais au minimum 10,
 * et au maximum 100).
 */
 
double guess_dist (strloc *str_loc, unsigned int nbpts)
{
    unsigned int max = MIN(nbpts, MAX(10, MIN(100, sqrt(nbpts))));
    double d, dmoy = 0;
    int i;

    for (i=0;i<max;i++)
      {
	
      };

    return (dmoy / max);
}

