/*
 * Definit tous les types de structure utilisees.
 * 
 */
 
#define MAX(a,b) ((a) < (b) ? (b) : (a))
#define MIN(a,b) ((a) > (b) ? (b) : (a))
#define SQR(a)   ((a) * (a))
#define TRUE  1
#define FALSE 0

#ifndef DBL_MAX
#define DBL_MAX 1.797693e+308
#endif

typedef struct {double x, y, z;} vect;
typedef struct {double x, y, z;} pt3d;
typedef struct {double x, y;} pt2d;

typedef struct {
    pt3d    pt;
    double  d;
} pt_dist;


typedef struct strlocn {
    unsigned short flag;    /* =0 pour un cube,
			     * =1 pour une feuille, 
			     * (pour vide: avec feuille=NULL)
			     */
    union {
	pt3d		*feuille;
	struct strlocn	*cube; /*
				* Pointe vers un tableau de 8 strloc.
				* *cube ====== cube[8], sauf que
				* c'est pas reserve a la declaration, 
				* mais avec un malloc.
				*/
    } kwa;
} strloc;


/*
 * str_mem == tableau dynamique (de pointeurs vers des pt3d).
 * 
 * Ce tableau sert de pile,  ou l'on entasse (et ne retire jamais
 * des pt3d*). C'est utilise pour rendre la liste des voisins d'un
 * point.
 * 
 * STR_MEM_STEP fixe la taille d'augmentation quand il est
 * necessaire d'augmenter la taille du tableau, ainsi que
 * la taille initiale.
 */

#define STR_MEM_STEP 128

typedef struct {
    pt3d	    **debut;	/* pointeur vers une zone de	*/
				/* memoire libre (de pointeurs  */
				/* vers des pt3d).		*/
    unsigned int    max;	/* Taille actuellement reservee	*/
    unsigned int    current;    /* Nombre de cases utilisees	*/
} str_mem;


/*
 * Liste des voisins d'un point.
 */

typedef struct list {
    pt3d *point;
    struct list *next;
} liste;

/*
 * Petit #define pour creer une liste (un element de la liste).
 */

#define NEW_LISTE(a) \
    { (a) = (liste *) malloc( sizeof(liste)); \
	if ((a) == NULL) { \
	    printf("Erreur d'allocation memoire dans NEWLISTE().\n"); \
	    exit(-1); \
	}; }

/*
 * Le maillage est un tableau de pointeurs
 * vers des listes (NULL si pas encore defini).
 * 
 * La liste est ordonnee, de sorte qu'un triangle
 * est forme par deux points consecutifs (ou le
 * premier + le dernier).
 * 
 * On determine ainsi rapidemment si un point
 * appartient a "la" frontiere ou pas, sans se
 * trainer une structure supplementaire.
 * 
 * nb represente le nombre de points.
 * first est un pointeur vers les coordonnees du premier point.
 * (donc vers le tableau des coordonnees).
 * 
 */

typedef struct {
			/* Parametres d'entree. */
    unsigned int nb;	    /* Nombre de points a mailler */
    pt3d *first;	    /* Pointeur vers le premier point,
			       Ca doit etre contigu. */
			/* Parametres de sortie. */
    struct paire {
	pt3d  *voisin;	    /* Si c'est NULL, le point est soit
				- pas du tout maille,  soit
				- entierement maille.
			       Sinon, c'est que le point est sur la
			       frontiere, et la valeur correspond
			       a un pointeur vers son suivant. */
	liste *voisins;	    /* Si c'est NULL, le point correspondant
			       n'est pas du tout maille.
			       Sinon, ca pointe vers une liste ORDONNEE
			       des voisins sur le maillage (points
			       connectes par des segments a ce point). */
	} *maille;	    /* pointeur vers un tableau de paires */
} maillage;
