/*
 * Fonctions d'affichage sous X.
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "structures.h"
#include "xdebug.h"

/*
 * C'est ici qu'on peut inserer la vue que l'on souhaite.
 */

static double view[] =
{ 13.63392132862818861838, 1.85738835123287882034, 5.50548497592829377822,
  -0.32763455759000703615, -13.77465842175621091314, 5.45852618050918358961,
  5.80111298839637967717, -5.14324420763159650960, -12.63084297342084738602,
  249.07007174319957698572, 882.66373099624854603462};


static Display *display;
static int screen;
static GC gc;
static Window win;

static XColor colours[] =
  {
    {0, 0xffff, 0xffff, 0xffff},    /* white */
    {0, 0x0000, 0x0000, 0x0000},    /* black */
    {0, 0xffff, 0x0000, 0x0000},    /* red   */
    {0, 0x0000, 0xffff, 0x0000},    /* green */
    {0, 0x0000, 0x0000, 0xffff},    /* blue  */
    {0, 0xffff, 0xffff, 0x0000},    /* yellow */
    {0, 0xffff, 0x0000, 0xffff},    /* violet */
    {0, 0x0000, 0xcccc, 0xcccc},    /* cyan  */
    {0, 0x8000, 0x8000, 0x8000},    /* grey */
    {0, 0x9000, 0x4000, 0x2000},    /* brown */
    {0, 0xffff, 0x6000, 0x0000},    /* orange */
  };

short coord_x_2D (pt3d *pt)
{
    int x = pt->x * view[0] +
	    pt->y * view[1] +
	    pt->z * view[2] +
    	    view[9];
    return (short) MAX (-32768, MIN (32767, x));
}

short coord_y_2D (pt3d *pt)
{
    int y = pt->x * view[3] +
	    pt->y * view[4] +
	    pt->z * view[5] +
	    view[10];
    return (short) MAX (-32768, MIN (32767, y));
}


void x_point ( pt3d *pt)
{
    XDrawPoint (display, win, gc, coord_x_2D (pt), coord_y_2D (pt));
}


void x_gros_point ( pt3d *pt)
{
    int x = coord_x_2D (pt);
    int y = coord_y_2D (pt);

    XDrawPoint (display, win, gc, x-1, y-1);
    XDrawPoint (display, win, gc, x-1, y);
    XDrawPoint (display, win, gc, x-1, y+1);
    XDrawPoint (display, win, gc, x, y-1);
    XDrawPoint (display, win, gc, x, y);
    XDrawPoint (display, win, gc, x, y+1);
    XDrawPoint (display, win, gc, x+1, y-1);
    XDrawPoint (display, win, gc, x+1, y);
    XDrawPoint (display, win, gc, x+1, y+1);
}

void x_segment ( pt3d *pt1, pt3d *pt2)
{
    XDrawLine ( display, win, gc, 
		coord_x_2D (pt1), coord_y_2D (pt1), 
		coord_x_2D (pt2), coord_y_2D (pt2));
}


void x_vecteur ( pt3d *pt1, vect *v)
{
    pt3d pt2;
    pt2.x = pt1->x + v->x;
    pt2.y = pt1->y + v->y;
    pt2.z = pt1->z + v->z;
    XDrawLine ( display, win, gc, 
		coord_x_2D (pt1), coord_y_2D (pt1), 
		coord_x_2D (&pt2), coord_y_2D (&pt2));
}


void x_cube ( pt_dist *cube)
{
    pt3d pt1, pt2;

    pt2.x = (pt1.x = cube->pt.x - cube->d / 2) + cube->d;
    pt2.y =  pt1.y = cube->pt.y - cube->d / 2;
    pt2.z =  pt1.z = cube->pt.z - cube->d / 2;
    
    x_segment ( &pt1,  &pt2);
    pt1.y += cube->d;
    pt2.y += cube->d;
    x_segment ( &pt1,  &pt2);
    pt1.z += cube->d;
    pt2.z += cube->d;
    x_segment ( &pt1,  &pt2);
    pt1.y -= cube->d;
    pt2.y -= cube->d;
    x_segment ( &pt1,  &pt2);

    pt2.x -= cube->d;
    pt2.z -= cube->d;

    x_segment ( &pt1,  &pt2);
    pt1.y += cube->d;
    pt2.y += cube->d;
    x_segment ( &pt1,  &pt2);
    pt1.x += cube->d;
    pt2.x += cube->d;
    x_segment ( &pt1,  &pt2);
    pt1.y -= cube->d;
    pt2.y -= cube->d;
    x_segment ( &pt1,  &pt2);

    pt1.z -= cube->d;
    pt1.y += cube->d;

    x_segment (	&pt1,  &pt2);
    pt1.z += cube->d;
    pt2.z += cube->d;
    x_segment ( &pt1,  &pt2);
    pt1.x -= cube->d;
    pt2.x -= cube->d;
    x_segment ( &pt1,  &pt2);
    pt1.z -= cube->d;
    pt2.z -= cube->d;
    x_segment ( &pt1,  &pt2);
}


void x_maillage (maillage *maille,
		 XSegment *tableau)
{
    unsigned int i, j;
    liste *list_elt;

    for (i=j=0; i<(maille->nb); i++)
    {
	list_elt = maille->maille[i].voisins;
	while (list_elt != NULL)
	{
	    if ( ((maille->first + i) < (list_elt->point)) &&
		 (maille->maille[i].voisin != list_elt->point) &&
		 (maille->maille[list_elt->point-maille->first].voisin != (maille->first + i)))
	    {
		tableau[j].x1 = coord_x_2D (maille->first + i);
		tableau[j].y1 =	coord_y_2D (maille->first + i);
		tableau[j].x2 = coord_x_2D (list_elt->point);
		tableau[j].y2 = coord_y_2D (list_elt->point);
		++j;
	    };
	    list_elt = list_elt->next;
	};
    };
    XDrawSegments( display, win, gc, tableau, j);
}


void x_frontiere (  maillage *maille, 
		    XSegment *tableau)
{
    unsigned int i, j;

    for (i=j=0; i < maille->nb; i++)
	if (maille->maille[i].voisin != NULL)
	{
	    tableau[j].x1 = coord_x_2D (maille->first + i);
	    tableau[j].y1 = coord_y_2D (maille->first + i);
	    tableau[j].x2 = coord_x_2D (maille->maille[i].voisin);
	    tableau[j].y2 = coord_y_2D (maille->maille[i].voisin);
	    ++j;
	};
    XDrawSegments( display, win, gc, tableau, j);
}


void x_points (  maillage *maille, 
		 XPoint *tableau)
{
    unsigned int i, j;

    for (i=j=0; i < maille->nb; i++)
    {
	tableau[j].x = coord_x_2D (maille->first + i);
	tableau[j].y = coord_y_2D (maille->first + i);
	++j;
    };
    XDrawPoints( display, win, gc, tableau, j, CoordModeOrigin);
}


void x_gros_points ( maillage *maille, 
		     XPoint *tableau)
{
    unsigned int i, j;
    unsigned short x, y;

    for (i=j=0; i < maille->nb; i++)
    {
	x = coord_x_2D (maille->first + i);
	y = coord_y_2D (maille->first + i);
	tableau[j].x = x-1; tableau[j].y = y-1; ++j;
	tableau[j].x = x-1; tableau[j].y = y;   ++j;
	tableau[j].x = x-1; tableau[j].y = y+1; ++j;
	tableau[j].x = x;   tableau[j].y = y-1; ++j;
	tableau[j].x = x;   tableau[j].y = y;   ++j;
	tableau[j].x = x;   tableau[j].y = y+1; ++j;
	tableau[j].x = x+1; tableau[j].y = y-1; ++j;
	tableau[j].x = x+1; tableau[j].y = y;   ++j;
	tableau[j].x = x+1; tableau[j].y = y+1; ++j;
    };
    XDrawPoints( display, win, gc, tableau, j, CoordModeOrigin);
}


void get_GC ( XFontStruct *font_info)
{
	unsigned long valuemask = 0;
	XGCValues values;
	unsigned int line_width = 1;
	int line_style = LineDoubleDash; /*LineOnOffDash;*/
	int cap_style = CapButt;
	int join_style = JoinMiter;
	int dash_offset = 0;
	static char dash_list[] = { 20, 40 };
	int list_length = sizeof (dash_list);
	gc = XCreateGC (display, win, valuemask, &values);

	XSetFont (display, gc, font_info->fid);
	XSetForeground (display, gc, WhitePixel(display,screen));
	XSetLineAttributes (display, gc, line_width, line_style,
			    cap_style, join_style);
	XSetDashes (display, gc, dash_offset, dash_list, list_length);
}


void load_font (XFontStruct **font_info)
{
	char *fontname = "9x15";

	if ((*font_info = XLoadQueryFont (display,fontname)) == NULL)
	{
		printf("stderr, basicwin: cannot open 9x15 font\n");
		exit(-1);
	};
}


void x_dessine (maillage *maillage_demo)
{
  unsigned int i;
  XWindowAttributes winattr;
  unsigned int width, height, display_width, display_height;
  int x = 0, y = 0;
  unsigned int border_width = 4;
  char *window_name = "Syl's Debug Window";
  char *icon_name = "Syl's X-Debug";
  XEvent report;
  XFontStruct *font_info;
  char *display_name = NULL;
  
  int boutton = 0;
  int trace;
  int oldx, oldy;

  int count;
  int Xnumberofcolors = 10;
  Colormap cmap;

  liste *list_elt;
  unsigned int nbseg = 0;
  unsigned int nbfront = 0;
  XSegment *tab_seg, *tab_fro;
  XPoint *tab_pt;

  if ((display = XOpenDisplay(display_name)) == NULL)
    {
      fprintf(stderr,"X-debug: cannot connect to X server %s\n",
	      XDisplayName (display_name));
      exit(-1);
    }
  
  screen = DefaultScreen(display);
  
  display_width  = DisplayWidth  (display, screen);
  display_height = DisplayHeight (display, screen);
  
  width = display_width/2;
  height = display_height/2;
  
  win = XCreateSimpleWindow (display, RootWindow(display, screen),
			     x, y, width, height, border_width,
			     BlackPixel(display, screen), 
			     BlackPixel(display, screen));
  
  XStoreName   (display, win, window_name);
  XSetIconName (display, win, icon_name);
  
  XSelectInput (display, win,
		ExposureMask      | KeyPressMask |
		ButtonPressMask   | ButtonReleaseMask |
		PointerMotionMask | StructureNotifyMask);
  
  load_font (&font_info);
  get_GC ( font_info);
  
  XMapWindow (display, win);
  
  if ( DisplayPlanes (display, screen) == 1)
    {
      printf ("Pour avoir des couleurs, faut un ecran couleur :-)\n");
      exit (-1);
    }
  else
    {
      cmap = DefaultColormap (display, screen);
      for (count=0; count<=Xnumberofcolors; ++count)
        {
	  colours[count].flags = DoRed | DoGreen | DoBlue;
	  XAllocColor ( display, cmap, &colours[count]);
	};
    };
  
  XSetLineAttributes (display, gc, 1, LineSolid, CapButt, JoinMiter);

  /*
   * Je calcule le nombre de segments sur la frontiere,
   * et a l'insterieur, et j'alloue les tableaux.
   */

    for (i=0; i<(maillage_demo->nb); i++)
    {
	if (maillage_demo->maille[i].voisin != NULL) ++nbfront;
	list_elt = maillage_demo->maille[i].voisins;
	while (list_elt != NULL)
	{
	    if ( ((maillage_demo->first + i) < (list_elt->point)) &&
		 (maillage_demo->maille[i].voisin != list_elt->point) &&
		 (maillage_demo->maille[list_elt->point-maillage_demo->first].voisin != (maillage_demo->first + i)))
		/*
		 * Trace 2 fois les memes segments,
		 * si on ne fait pas ce test.
		 */
	    ++nbseg;
	    list_elt = list_elt->next;
	};
    };

    tab_seg = malloc ( nbseg * sizeof(XSegment));
    tab_fro = malloc ( nbfront * sizeof(XSegment));
    tab_pt  = malloc ( 9 * maillage_demo->nb * sizeof(XPoint));
		    /* "* 9", c'est pour les gros points. */
    if ((tab_seg == NULL) || (tab_fro == NULL) || (tab_pt == NULL)) {
	printf("Erreur d'allocation memoire dans x_dessinne().\n");
	exit(-1);
    };


  /* Boucle d'affichage. */

    XNextEvent(display, &report);
      oldx = report.xbutton.x;
      oldy = report.xbutton.y;
      trace = 1;
        
  while (TRUE)
  {

    XNextEvent(display, &report);

    do {
      switch (report.type)
	{
	case KeyPress:
	  {
	    printf("Coordonnnes de la vue:\n");
	    printf("{ %.20f, %.20f, %.20f,\n", view[0], view[1], view[2]);
	    printf("  %.20f, %.20f, %.20f,\n", view[3], view[4], view[5]);
	    printf("  %.20f, %.20f, %.20f,\n", view[6], view[7], view[8]);
	    printf("  %.20f, %.20f};\n", view[9], view[10]);
	    break;
	  };
	case ButtonPress:
	  {
	    boutton = report.xbutton.button;
	    break;
	  };
	case ButtonRelease:
	  {
	    boutton = 0;
	    break;
	  };
	case MotionNotify:
	  if (boutton != 0) trace = 1;
	  switch (boutton)
	    {
	    case Button1:
	      {
		double tetay = (report.xbutton.x - oldx) / 360.0 * 3.1415;
		double tetax = (report.xbutton.y - oldy) / 360.0 * 3.1415;
		double new_view[9];
		
		/* Rotation par rapport a l'axe (y). */
		
		new_view[0] = view[0] * cos(tetay) - view[6] * sin(tetay);
		new_view[1] = view[1] * cos(tetay) - view[7] * sin(tetay);
		new_view[2] = view[2] * cos(tetay) - view[8] * sin(tetay);
		new_view[3] = view[3];
		new_view[4] = view[4];
		new_view[5] = view[5];
		new_view[6] = view[0] * sin(tetay) + view[6] * cos(tetay);
		new_view[7] = view[1] * sin(tetay) + view[7] * cos(tetay);
		new_view[8] = view[2] * sin(tetay) + view[8] * cos(tetay);
		
		/* Rotation par rapport a l'axe (x). */
		
		view[0] = new_view[0];
		view[1] = new_view[1];
		view[2] = new_view[2];
		view[3] = new_view[3] * cos(tetax) - new_view[6] * sin(tetax);
		view[4] = new_view[4] * cos(tetax) - new_view[7] * sin(tetax);
		view[5] = new_view[5] * cos(tetax) - new_view[8] * sin(tetax);
		view[6] = new_view[3] * sin(tetax) + new_view[6] * cos(tetax);
		view[7] = new_view[4] * sin(tetax) + new_view[7] * cos(tetax);
		view[8] = new_view[5] * sin(tetax) + new_view[8] * cos(tetax);
		break;
	      }
	    case Button2:
	      {
		double tetaz = (report.xbutton.x - oldx) / 360.0 * 3.1415;
		double new_view[11];
		
		/* Je ne considere que le max de ces 2 actions. */
		
		if ( abs(report.xbutton.y - oldy) > abs(report.xbutton.x - oldx))
		  {/* Zoooooooom */

		    double lambda = (1.0 + (report.xbutton.y - oldy) / 100.0);		    
		    XGetWindowAttributes(display, win, &winattr);
		    for (i=0; i<9; i++)
			view[i] *= lambda;
		    view[9]  = (1-lambda) * winattr.width / 2 + lambda * view[9];
		    view[10] = (1-lambda) * winattr.height / 2 + lambda * view[10];
		  }
		else
		  { /* Rotation par rapport a l'axe (z). */
		    
		    XGetWindowAttributes(display, win, &winattr);

		    new_view[0] = view[0] * cos(tetaz) - view[3] * sin(tetaz);
		    new_view[1] = view[1] * cos(tetaz) - view[4] * sin(tetaz);
		    new_view[2] = view[2] * cos(tetaz) - view[5] * sin(tetaz);
		    
		    new_view[3] = view[0] * sin(tetaz) + view[3] * cos(tetaz);
		    new_view[4] = view[1] * sin(tetaz) + view[4] * cos(tetaz);
		    new_view[5] = view[2] * sin(tetaz) + view[5] * cos(tetaz);
		    
		    new_view[9] =   winattr.width / 2 + 
				    cos(tetaz) * (view[9]  - winattr.width / 2) -
				    sin(tetaz) * (view[10] - winattr.height / 2);
		    new_view[10] =  winattr.height / 2 + 
				    sin(tetaz) * (view[9]  - winattr.width / 2) +
				    cos(tetaz) * (view[10] - winattr.height / 2);
		    
		    view[0] = new_view[0];
		    view[1] = new_view[1];
		    view[2] = new_view[2];
		    
		    view[3] = new_view[3];
		    view[4] = new_view[4];
		    view[5] = new_view[5];

		    view[9] = new_view[9];
		    view[10] = new_view[10];
		  };
		break;
	      }
	    case Button3:
	      {
		/* Translation */
		view[9]  += report.xbutton.x - oldx;
		view[10] += report.xbutton.y - oldy;
		break;
	      }
	    default:
	      break;
	    };
	default:
	  break;
	};
      oldx = report.xbutton.x;
      oldy = report.xbutton.y;

    } while (XCheckWindowEvent ( display, win,
				 ExposureMask      | KeyPressMask |
    				 ButtonPressMask   | ButtonReleaseMask |
				 PointerMotionMask | StructureNotifyMask, 
				 &report));

      /* Sous-bloc affichant tout ce qu'on veut. */

	  if (trace != 0)
	  {
	      XClearWindow	(display, win);
	      XSetForeground	(display, gc, colours[5].pixel);
	      x_frontiere	( maillage_demo, tab_fro);
	      XSetForeground	(display, gc, colours[4].pixel);
	      x_maillage	( maillage_demo, tab_seg);
	      XSetForeground	(display, gc, colours[5].pixel);
	      x_frontiere	( maillage_demo, tab_fro);
	      XSetForeground	(display, gc, colours[6].pixel);
	      x_points	        ( maillage_demo, tab_pt);
	      XSetForeground	(display, gc, WhitePixel (display, screen));
	      trace = 0;
	  };
    };
    free (tab_seg);
    free (tab_fro);
    free (tab_pt);
}
